@@ -333,8 +333,10 @@ wvDecodeComplex (wvParseStruct * ps)
 {
     U32 piececount = 0, i, j, spiece = 0;
     U32 beginfc, endfc;
+	U32 stream_size;
     U32 begincp, endcp;
-    U8 chartype;
+    int ichartype;
+    U8  chartype;
     U16 eachchar;
     U32 para_fcFirst, para_fcLim = 0xffffffffL;
     U32 dummy, nextpara_fcLim = 0xffffffffL;
@@ -365,6 +367,9 @@ wvDecodeComplex (wvParseStruct * ps)
     U32 *posBKL;
     U32 bkl_intervals;
 
+    external_wvReleasePAPX_FKP ();
+    external_wvReleaseCHPX_FKP ();
+
     /*dop */
     wvGetDOP (wvQuerySupported (&ps->fib, NULL), &ps->dop, ps->fib.fcDop,
 	      ps->fib.lcbDop, ps->tablefd);
@@ -399,7 +404,8 @@ encoded into the first 22 bytes.
     wvGetBKF_PLCF (&bkf, &posBKF, &bkf_intervals, ps->fib.fcPlcfAtnbkf,
 		   ps->fib.lcbPlcfAtnbkf, ps->tablefd);
     wvGetBKL_PLCF (&bkl, &posBKL, &bkl_intervals, ps->fib.fcPlcfAtnbkl,
-		   ps->fib.lcbPlcfAtnbkl, ps->tablefd);
+           ps->fib.lcbPlcfAtnbkl, ps->fib.fcPlcfAtnbkf, ps->fib.lcbPlcfAtnbkf,
+           ps->tablefd);
 
     /*we will need the stylesheet to do anything useful with layout and look */
     wvGetSTSH (&ps->stsh, ps->fib.fcStshf, ps->fib.lcbStshf, ps->tablefd);
@@ -507,14 +513,27 @@ encoded into the first 22 bytes.
 
     wvHandleDocument (ps, DOCBEGIN);
 
+	/*get stream size for bounds checking*/
+	stream_size = wvStream_size(ps->mainfd);
 
     /*for each piece */
     for (piececount = 0; piececount < ps->clx.nopcd; piececount++)
       {
-	  chartype =
+	  ichartype =
 	      wvGetPieceBoundsFC (&beginfc, &endfc, &ps->clx, piececount);
+	  if(ichartype==-1)
+		  break;
+	  chartype = (U8) ichartype;
+	  /*lvm007@aha.ru fix antiloop: check stream size */
+	  if(beginfc>stream_size || endfc>stream_size){
+		  wvError (
+		   ("Piece Bounds out of range!, its a disaster\n"));
+		  continue;
+	  }
 	  wvStream_goto (ps->mainfd, beginfc);
-	  wvGetPieceBoundsCP (&begincp, &endcp, &ps->clx, piececount);
+	  /*lvm007@aha.ru fix antiloop fix*/
+	  if(wvGetPieceBoundsCP (&begincp, &endcp, &ps->clx, piececount)==-1)
+		  break;
 	  wvTrace (
 		   ("piece begins at %x and ends just before %x. the char end is %x\n",
 		    beginfc, endfc, char_fcLim));
@@ -752,6 +771,11 @@ encoded into the first 22 bytes.
 			      }
 
 			}
+		      else{
+  			 /* lvm007@aha.ru fix: if currentfc>fcFirst but CHARPROP's changed look examples/charprops.doc for decode_simple*/
+			 if(char_fcFirst< j)
+				char_fcFirst = j;
+		       }
 		  }
 
 		if (j == char_fcFirst)
@@ -925,7 +949,7 @@ wvGetComplexSEP (wvVersion ver, SEP * sep, U32 cpiece, STSH * stsh, CLX * clx)
 #endif
 	  RetSprm =
 	      wvApplySprmFromBucket (ver,
-				     wvGetrgsprmPrm (clx->pcd[cpiece].prm.
+				     (U16) wvGetrgsprmPrm ( (U16) clx->pcd[cpiece].prm.
 						     para.var1.isprm), NULL,
 				     NULL, sep, stsh, pointer, &pos, NULL);
 	  if (RetSprm.sgc == sgcSep)
@@ -952,7 +976,7 @@ wvGetComplexSEP (wvVersion ver, SEP * sep, U32 cpiece, STSH * stsh, CLX * clx)
 		else
 		  {
 		      sprm = bread_8ubit (clx->grpprl[index] + i, &i);
-		      sprm = (U8) wvGetrgsprmWord6 (sprm);
+		      sprm = (U8) wvGetrgsprmWord6 ( (U8) sprm);
 		  }
 		pointer = clx->grpprl[index] + i;
 		RetSprm =
@@ -1000,7 +1024,7 @@ wvAssembleComplexPAP (wvVersion ver, PAP * apap, U32 cpiece, STSH * stsh,
 #endif
 	  RetSprm =
 	      wvApplySprmFromBucket (ver,
-				     wvGetrgsprmPrm (clx->pcd[cpiece].prm.
+				     (U16) wvGetrgsprmPrm ( (U16) clx->pcd[cpiece].prm.
 						     para.var1.isprm), apap,
 				     NULL, NULL, stsh, pointer, &pos, data);
 	  if (RetSprm.sgc == sgcPara)
@@ -1060,7 +1084,7 @@ wvAssembleComplexCHP (wvVersion ver, CHP * achp, U32 cpiece, STSH * stsh,
 #endif
 	  RetSprm =
 	      wvApplySprmFromBucket (ver,
-				     wvGetrgsprmPrm (clx->pcd[cpiece].prm.
+				     (U16) wvGetrgsprmPrm ( (U16) clx->pcd[cpiece].prm.
 						     para.var1.isprm), NULL,
 				     achp, NULL, stsh, pointer, &pos, NULL);
 	  if (RetSprm.sgc == sgcChp)
