@@ -25,8 +25,10 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
     CHP achp;
     U32 piececount = 0, i, j = 0, spiece;
     U32 beginfc, endfc;
+	U32 stream_size;
     U32 begincp, endcp;
-    U8 chartype;
+    int ichartype;
+    U8  chartype;
     U16 eachchar;
     U32 para_fcFirst, para_fcLim = 0xffffffff;
     U32 dummy, nextpara_fcLim = 0xffffffff;
@@ -59,6 +61,9 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
     U32 bkd_intervals;
     wvVersion ver;
 
+    external_wvReleasePAPX_FKP ();
+    external_wvReleaseCHPX_FKP ();
+
     ver = wvQuerySupported (&ps->fib, NULL);
 
 
@@ -103,7 +108,8 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
     wvGetBKF_PLCF (&bkf, &posBKF, &bkf_intervals, ps->fib.fcPlcfAtnbkf,
 		   ps->fib.lcbPlcfAtnbkf, ps->tablefd);
     wvGetBKL_PLCF (&bkl, &posBKL, &bkl_intervals, ps->fib.fcPlcfAtnbkl,
-		   ps->fib.lcbPlcfAtnbkl, ps->tablefd);
+           ps->fib.lcbPlcfAtnbkl,ps->fib.fcPlcfAtnbkf, ps->fib.lcbPlcfAtnbkf,
+           ps->tablefd);
 
 
     /* get font list */
@@ -235,15 +241,31 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
     wvHandleDocument (ps, DOCBEGIN);
 
 
+	/*get stream size for bounds checking*/
+	stream_size = wvStream_size(ps->mainfd);
 
     /*for each piece */
     for (piececount = 0; piececount < ps->clx.nopcd; piececount++)
       {
-	  chartype =
+	  ichartype =
 	      wvGetPieceBoundsFC (&beginfc, &endfc, &ps->clx, piececount);
+	  if(ichartype==-1)
+		  break;
+	  chartype = (U8) ichartype;
+	  /*lvm007@aha.ru fix antiloop: check stream size */
+	  if(beginfc>stream_size || endfc>stream_size){
+		  wvError (
+		   ("Piece Bounds out of range!, its a disaster\n"));
+		  continue;
+	  }
+
 	  wvStream_goto (ps->mainfd, beginfc);
+
 	  wvTrace (("SEEK %x\n", beginfc));
-	  wvGetPieceBoundsCP (&begincp, &endcp, &ps->clx, piececount);
+
+	  /*lvm007@aha.ru fix antiloop fix*/
+	  if(wvGetPieceBoundsCP (&begincp, &endcp, &ps->clx, piececount)==-1)
+		  break;
 
 	  /*
 	     text that is not in the same piece is not guaranteed to have the same properties as
@@ -425,6 +447,11 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
 			      }
 
 			}
+		      else{
+  			/* lvm007@aha.ru fix: if currentfc>fcFirst but CHARPROP's changed look examples/charprops.doc*/
+			if(char_fcFirst< j)
+				char_fcFirst = j;
+		       }
 		  }
 
 		if (j == char_fcFirst)
@@ -509,11 +536,23 @@ wvDecodeSimple (wvParseStruct * ps, subdocument whichdoc)
     wvFree (posAtrd);
     wvFree (atrd);
 
-    wvReleasePAPX_FKP (&para_fkp);
-    wvReleaseCHPX_FKP (&char_fkp);
+    internal_wvReleasePAPX_FKP (&para_fkp);
+    internal_wvReleaseCHPX_FKP (&char_fkp);
     wvHandleDocument (ps, DOCEND);
     wvFree (posSedx);
     wvFree (sed);
+    wvFree(bkd);
+    wvFree(posBKD);
+    wvFree(ftxbx);
+    wvFree(txbxTxt); 
+#if defined(WIN32)
+   wvReleasePAPX_FKP (&para_fkp);
+   wvReleaseCHPX_FKP (&char_fkp);   
+
+   external_wvReleasePAPX_FKP ();
+   external_wvReleaseCHPX_FKP();
+#endif
+
 
     wvFree (ps->liststartnos);
     wvFree (ps->listnfcs);
