@@ -37,6 +37,17 @@
 #include <ctype.h>
 #include <string.h>
 
+#ifdef __MINGW32__ /* incomplete defines on mingw */
+
+#define _S_IXGRP        _S_IEXEC
+#define _S_IWGRP        _S_IWRITE
+#define _S_IRGRP        _S_IREAD
+
+#define S_IXGRP         _S_IXGRP
+#define S_IWGRP         _S_IWGRP
+#define S_IRGRP         _S_IRGRP
+
+#endif
  
 #ifdef HAVE_MMAP
 /*lvm007@aha.ru fix*/
@@ -1751,8 +1762,8 @@ ms_ole_open_vfs (MsOle **fs, const char *name,
 		f->mem = g_new (guint8, BB_BLOCK_SIZE);
 
 		if (!f->mem ||
-		    f->syswrap->read (file, f->mem, (BB_BLOCK_SIZE == -1),
-				      f->syswrap->closure)) {
+		    f->syswrap->read (file, f->mem, BB_BLOCK_SIZE,
+				      f->syswrap->closure) == -1) {
 			g_warning ("Error reading header\n");
 			f->syswrap->close (file, f->syswrap->closure) ;
 			g_free (f);
